--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2012  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=2.9991;

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end
local DM=DuckMod[TV];
local _;

-- Helper colours
DM.Color = {
	White="|cFFFFFFFF",
	Green="|cFF00FF00",
	Red="|cFFFF0000",
	lBlue="|cFF6060FF",				-- Light blue
	hBlue="|cFEA0A0FF",				-- Highlight blue
	Yellow="|cFFFFFF00",
};

DM.RenderQueue={};
DM.Render={
	Class={
		-- "=" denotes the named class
		-- "+" denotes included entities
		HaveContainer = 0x00000001, -- - - - + + - - - - - - - - - -
		UIObject	  = 0x00000002, -- + + + + + + + + + + + + + + +
		FontInstance  = 0x00000004, -- = - - + - - + - - + + - - - -
		Region        = 0x00000008, -- - = + + + + + + + + + + + + +
		LayeredRegion = 0x00000010, -- - - = + + - - - - - - - - - -
		FontString    = 0x00000020, -- - - - = - - - - - - - - - - -
		Texture       = 0x00000040, -- - - - - = - - - - - - - - - -
		Frame         = 0x00000080, -- - - - - - = + + + + + + + + +
		EditBox       = 0x00000100, -- - - - - - - = - - - - - - - -
		Button        = 0x00000200, -- - - - - - - - = - - - - + + -
		SimpleHTML    = 0x00000400, -- - - - - - - - - = - - - - - -
		ScrMsgFrame   = 0x00000800, -- - - - - - - - - - = - - - - -
		SwapToLink    = 0x00001000, -- - - - - - - - - - - + - - - -
		DeadEditBox   = 0x00002000, -- - - - - - - - - - - = - - - -
		NoAutoEvent   = 0x00004000, -- - - - - - - - - - - - - - - -
		List          = 0x00008000, -- - - - - - - - - - - - = - - -
		Extended      = 0x00010000, -- - - - - - - - - - - - - + + -
		ChildText     = 0x00020000, -- - - - - - - - - - - - - + + -
		CheckBox      = 0x00040000, -- - - - - - - - - - - - - = + -
		RadioButton   = 0x00080000, -- - - - - - - - - - - - - - = -
		Bar           = 0x00100000, -- - - - - - - - - - - - - - - =

		FontString_   = 0x0000003F, -- Full FontString
		Texture_	  = 0x0000005B, -- Full Texture
		iTexture_     = 0x000042DA, -- Full button and texture with ItemButtonTemplate, no auto event
		Frame_        = 0x0000008A, -- Full Frame
		EditBox_      = 0x0000018E, -- Full EditBox
		Button_       = 0x0000028A, -- Full Button
		TextLink_     = 0x0000083F, -- FontString + SwapToLink
		SimpleHTML_   = 0x0000148E, -- Full SimpleHTML + swap-flag
		ScrMsgFrame_  = 0x0000088E, -- Full ScrollingMessageFrame
		DeadEditBox_  = 0x0000318E, -- Full editBox with no interaction
		List_         = 0x0000808A, -- Full DM_01 list
		CheckBox_     = 0x0007028A,	-- Full native checkbox extended
		RadioButton_  = 0x000F028A,	-- Full native checkbox extended with radiobutton
		Bar_          = 0x0010008A, -- A graphic bar
	},
};
local ren=DM.Render;


function ren:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Render-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod rendering UI frame",1); return; end
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-RenderWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod rendering WM frame",1); return; end
	end
	self.TheFrame:SetScript("OnUpdate",self.Tick);
	self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
end

function ren:Render(frame,sectiondata)
	if (not DM.RenderQueue) then DM.RenderQueue={}; end
	table.insert(DM.RenderQueue,{frame=frame,sectiondata=sectiondata});
end

-- Check the return value for success. "nil" is fail. "true" is victory.
-- If "nil", retry next update-frame.
function ren:Render_S(frame,sectiondata)
	if (ren.Canvas) then return nil; end	-- Could not render at this time
	if (type(sectiondata)=="table") then			-- SECTIONDATA
	elseif (type(sectiondata)=="string") then		-- SECTIONDATA.Data
		sectiondata={
			Data=sectiondata,
			Keys="",
			Description="Anonymous data",
			Meta={
				Server="",
				Location="",
			},
		};
	else return; end

	ren.Canvas=frame;
	ren.Canvas.SECTIONDATA=sectiondata;
	if (not frame.WnRenderData) then
		frame.WnRenderData={};
		frame.WnRenderData.Widget={};
		frame.WnRenderData.Focus={};
		frame.WnRenderData.WnContext={};
		frame.GoTo=ren.SetCanvasFocus;
		frame.GetForm=ren.GetContextData;
		frame.GetObject=ren.GetObject;
	end

	DM.RenderWnCounter=0;
	DM.RenderWnRemaining=sectiondata.Data;
	if (ren.Canvas:GetParent().SetVerticalScroll) then ren.Canvas:GetParent():SetVerticalScroll(0); end
	if (ren:Render_2(ren.Canvas,DM.RenderWnRemaining)) then return true; end
	-- Fail, so do fail-stuff
	ren.Canvas=nil;
	DM.RenderWnCounter=1000;
	return nil;
end

local CallbackCount=3;		-- do callback at the count of three

function ren:Render_2(canvas,remaining)
	DM.RenderWnCounter=DM.RenderWnCounter+1;
	canvas.WnRenderData.LockH=0;			-- Far left
--DM:Chat("Rendering...",1,0,1);
	-- Do blocks
	local entryname,param,data;
--	entryname,param,data,remaining=ren:PullSection(remaining);
	entryname,param,remaining=ren:PullSection(remaining);
	if (entryname~="<wn>") then return; end			-- Wrong data-type waiting
	ren:InitPage(canvas);	-- This will work, as the FontString widget will be set properly before a new redraw
	while (remaining~="") do
		entryname,param,data,remaining=ren:PullSection(remaining);
		if (entryname=="<meta>") then
--DM:Chat("Setting up meta...",1,0,1);
			local remM=data;						-- Do the meta data
			while (remM~="") do
				entryname,param,data,remM=ren:PullSection(remM);
				if (entryname=="<elements>") then
					local remE=data;				-- Do the elements data
					while (remE~="") do
						entryname,param,data,remE=ren:PullSection(remE);
						-- Overlay parameters
						canvas.WnRenderData.Meta["<elements>"][entryname]=ren:Mix(param,canvas.WnRenderData.Meta["<elements>"][entryname]);
						ren:SetInterface(canvas,entryname,canvas.WnRenderData.Meta["<elements>"][entryname].Class)
					end
				elseif (entryname=="<font>") then
					local name="Default";
					if (param.name) then name=param.name; end
					param.name=nil;
					canvas.WnRenderData.Meta.font[name]=DM:CopyTable(param,canvas.WnRenderData.Meta.font[name]);
--for p,d in pairs(canvas.WnRenderData.Meta.font[name]) do print(name,p,d); end
				elseif (entryname=="<package>") then
					if (WoWnet.Scheme[data]) then
						-- load the scheme
					end
				else
					canvas.WnRenderData.Meta[entryname]=DM:CopyTable(param,canvas.WnRenderData.Meta[entryname]);
				end
			end
		elseif (entryname=="<body>") then
			if (not canvas.MeasureFontString) then
				canvas.MeasureFontString=ren:GetWidget(canvas,"<stext>",canvas.WnRenderData.Meta["<elements>"]["<stext>"],nil);
				if (not canvas.MeasureFontString) then
					DM:Chat("Could not create a DuckMod measurement frame",1);
					return;
				end
				canvas.MeasureFontString.SPECIALNOREUSE=true;
			end
--DM:Chat("Rendering body...",1,0,1);
			-- Set the page standards from meta
			if (canvas.WnRenderData.Meta["<page>"]) then
				if (canvas.WnRenderData.Meta["<page>"].background) then
					local bgwidget=ren:GetWidget(canvas,"<img>",canvas.WnRenderData.Meta["<elements>"]["<img>"],nil);
					bgwidget:SetDrawLayer("BACKGROUND");
					bgwidget:ClearAllPoints();
					bgwidget:SetPoint("TOPLEFT",canvas,"TOPLEFT");
					bgwidget:SetPoint("BOTTOMRIGHT",canvas,"BOTTOMRIGHT");
					local r,g,b,a=ren:SplitColor(canvas.WnRenderData.Meta["<page>"].background);	-- Try to make RGBA
					if (r) then bgwidget:SetTexture(r,g,b,a);							-- Single color
					else bgwidget:SetTexture(canvas.WnRenderData.Meta["<page>"].background); end		-- Image-file
					bgwidget:Show();
				end
			end
			-- Do the contents from data
			local remB=data;						-- Do the body data
			while (remB~="") do
				entryname,param,data,remB=ren:PullSection(remB);
--print(data);
				if (not entryname) then
					DM:Chat("Unknown tag: ?",1);
				elseif (entryname=="<set>") then
					ren:SetContext(canvas,param);	-- Non-visual elements
				elseif (entryname=="<focus>") then
					if (param.id) then canvas.WnRenderData.Focus[param.id]=canvas.WnRenderData.Bottom; end
				elseif (not canvas.WnRenderData.Meta["<elements>"][entryname]) then
					DM:Chat("Unknown tag: "..entryname,1);
				else
--print(entryname);
					local e=canvas.WnRenderData.Meta["<elements>"][entryname];
--print(e.IF.SetData);
					local swap=ren:Flag(canvas,e,"SwapToLink");
					-- Visual elements
					if (not e) then DM:Chat("Missing element <"..entryname..">",1); return; end
					if (not data) then data=e.defaultdata; end;
					-- * Make parameters
					local mix=ren:Mix(param,e);
					-- * Create a widget, or retrieve an old from a stack.
					local widget=ren:GetWidget(canvas,entryname,e,mix);
					-- * Set provided data
					local r,g,b,a=ren:SplitColor(data);	-- Try to make an RGBA
					if (not r or (entryname~="<img>" and entryname~="<hr>")) then
						if (ren:Flag(canvas,e,"FontInstance")) then
							ren:SetFont(canvas,widget,mix);
						elseif (ren:Flag(canvas,e,"ChildText")) then
							ren:SetFont(canvas,_G[widget:GetName().."Text"],mix);
						end
						widget[e.IF.SetData](widget,ren:toASCII(data));	-- Simulate colon
					else
						widget[e.IF.SetData](widget,r,g,b,a);	-- Simulate colon
					end
					-- * Set it for placeholder if needed
					if (swap) then
						ren:InitWidget(canvas.MeasureFontString);
						ren:CopyTextAttributes(widget,canvas.MeasureFontString);
						canvas.MeasureFontString:SetText(ren:toASCII(data));
					end
					-- * Anchor and size
					ren:SetPosition(canvas,widget,mix,swap);
					-- * set data for lists after size has been set
					if (ren:Flag(canvas,e,"List")) then
						widget[e.IF.SetData](widget,ren:toASCII(data));	-- Simulate colon
					end
					-- * Set it for the actual widget if needed
					if (swap) then
						widget[e.IF.SetData](widget,ren:toASCII(data)); -- Simulate colon
					end
--[[					if (widget.SetBackdrop) then
						widget:SetBackdrop({
							bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",
							edgeFile=nil,
							tile=false,
							tileSize=1,
							edgeSize=0,
							insets={
								left=-2,
								right=-2,
								top=-2,
								bottom=-2,
							},
						});
					end
]]
					-- * Show it
					widget:Show();
					canvas.WnRenderData.LastWidget=widget;
					-- * Callback for widget
					if (mix and mix["callback"] and DM.RenderWnCounter==CallbackCount) then
						local p=DM:CopyTable(mix) or {};
						local c=DM:CopyTable(canvas.WnRenderData.WnContext) or {};
						DM:StoreByText(p["callback"],canvas,widget,p,c,data);
					end
				end
			end
			canvas:SetHeight(canvas.WnRenderData.Bottom+1);
		else
			-- Unknown block
		end
	end
	DM.RenderWnCounter=DM.RenderWnCounter+1;

	-- Finalise the page
	if (DM.RenderWnCounter==4) then
		ren.Canvas=nil;						-- Done with this canvas
	end
--	if (DM.RenderWnCounter==2) then
--		DM.RenderWnCounter=4;
--		ren.Canvas=nil;						-- Done with this canvas
--	end
	return true;
end


function ren:Tick(elapsed)
	if (DM.RenderWnCounter==2) then
		ren:Render_2(ren.Canvas,DM.RenderWnRemaining);
		table.remove(DM.RenderQueue,1);
	elseif (next(DM.RenderQueue)) then
		ren:Render_S(DM.RenderQueue[1].frame,DM.RenderQueue[1].sectiondata);
		
	end
end



-- Deviations to the default set-up of parameters are done by table overlay.
ren.DefaultMeta={
	-- Create the building-blocks
	Package="",											-- A package reference
	["<page>"]={
		background="#0000",								-- #AARRGGBB or filename
	},
	font={			-- EXCEPTION: Not bracketed, as it is not a tag-to-entry table
		Default={
			file=[[Fonts\FRIZQT__.TTF]],
			size="14",
			color="#ffff",
			flags="",
		},
		Medium={
			size="22",
		},
		Big={
			size="28",
		},
		Edit={
			color="#FAAA",
		},
	},
	-- Create the elements
	["<elements>"]={
		["<text>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			Class=ren.Class.FontString_,
		},
		["<h1>"]={
			widget="FontString",				-- the widget
			font="Big",
			pad=3,
			Class=ren.Class.FontString_,
		},
		["<h2>"]={
			widget="FontString",				-- the widget
			font="Medium",
			pad=3,
			Class=ren.Class.FontString_,
		},
		["<img>"]={
			widget="Texture",					-- the widget
			pad=3,
			width=32,
			height=32,
			defaultdata=[[Interface\Icons\INV_Misc_QuestionMark]],
			Class=ren.Class.Texture_,
		},
		["<edit>"]={
			widget="EditBox",					-- the widget
			inheritsys="DuckMod_EditText_01",	-- Inherit a frame
			width=200,
			height=25,
			font="Edit",
			pad=3,
			Class=ren.Class.EditBox_,
		},
		["<button>"]={
			widget="Button",					-- the widget
			inheritsys="UIPanelButtonTemplate",
			width=125,
			height=22,
			pad=3,
			Class=ren.Class.Button_,
		},
		["<list>"]={
			widget="Frame",						-- the widget
			inheritsys="DuckMod_ListBox_01",
			width=200,
			height=100,
			pad=3,
			Class=ren.Class.List_,
		},
		["<hr>"]={
			widget="Texture",					-- the widget
			pad=3,
			width="30%",
			height=1,
			defaultdata="#ffff",
			Class=ren.Class.Texture_,
		},
		["<frame>"]={
			widget="Frame",						-- the widget
			width=128,
			height=32,
			pad=3,
			Class=ren.Class.Frame_,
		},

		["<none>"]={							-- dummy
			Class=0,
		},
		["<stext>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			Class=ren.Class.FontString_,
		},
		["<ltext>"]={
			font="Default",
			pad=3,
			widget="EditBox",					-- the widget
			Class=ren.Class.DeadEditBox_,
		},
		["<img_i>"]={
			widget="Button",					-- the widget
			inheritsys="ItemButtonTemplate",
			pad=3,
			Class=ren.Class.iTexture_,
		},
		["<mledit>"]={
			widget="EditBox",					-- the widget
			inheritsys="DuckMod_EditText_01",	-- Inherit a frame
			width=200,
			height=100,
			font="Edit",
			pad=3,
			Class=ren.Class.EditBox_,
		},
		["<check>"]={
			widget="CheckButton",				-- the widget
			inheritsys="UICheckButtonTemplate",
			width=100,
			height=25,
			pad=3,
			Class=ren.Class.CheckBox_,
		},
		["<radio>"]={
			widget="CheckButton",				-- the widget
			inheritsys="UICheckButtonTemplate",
			width=100,
			height=25,
			pad=3,
			Class=ren.Class.RadioButton_,
		},
		["<bar>"]={
			widget="Frame",						-- the widget
			width=100,
			height=14,
			pad=3,
			Class=ren.Class.Bar_,
		},
	},
};

-- Each canvas (frame) will have the following additional info:
-- canvas
--	WnRenderData
--	   Widget	 - Storage for widget objects
--	   Meta	   - Meta for this page. That is: combined default and provided.
--	   LastWidget - Last rendered widget
--	   Bottom	 - Current end-of-page
--	   WnContext  - Current page context (<set>)
--	   LockH	  - The current LockH while rendering

-- Since each canvas keeps its own registry, multiple rendered pages is supported.
function ren:InitPage(canvas)
	canvas.WnRenderData.Meta=DM:CopyTable(ren.DefaultMeta);	-- Set default meta
	canvas.WnRenderData.LastWidget=nil;				-- nil is the container frame, and true only for the first widget
	canvas.WnRenderData.Bottom=0;					-- Bottom of reendered page
	canvas.WnRenderData.LockH=0;					-- Last locked horizontal
	-- START: v2.9903
	canvas:SetScript("OnMouseDown",ren.CanvasClicked);
--	if (not canvas.WnLinkMenu) then canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",canvas,"UIDropDownMenuTemplate"); end
	--if (not canvas.WnLinkMenu) then
--		canvas.WnLinkMenuAnchor=CreateFrame("Frame","WnLinkerMenuDisplayPositioning",UIParent);
--		canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",canvas.WnLinkMenuAnchor,"UIDropDownMenuTemplate");
		--canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",UIParent,"UIDropDownMenuTemplate");
	--end
	-- END
	wipe(canvas.WnRenderData.WnContext);			-- Current run-length context
	wipe(canvas.WnRenderData.Focus);				-- Clear foci
	-- Clear all used elements
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
			wwTable.Used=nil;						-- New page, so all unused
			ren:InitWidget(wwTable.widget);
		end
	end

--DM:Chat("Setting default interfaces...",1,0,1);
	-- Set interface for all defaults
	for nElement,dElement in pairs(canvas.WnRenderData.Meta["<elements>"]) do
		self:SetInterface(canvas,nElement,dElement.Class)
	end
end

-- v2.9903
function ren.CanvasClicked(frame,button,down)
--print("canvas",button)
--	if (button~="LeftButton") then return; end
	for element,eTable in pairs(frame.WnRenderData.Widget) do
		for name,nTable in pairs(eTable) do
			if (nTable.widget and nTable.widget.IsMouseOver) then
				if (nTable.widget:IsMouseOver()) then		-- We're hovering this widget
					if (nTable.widget.WnParam and nTable.widget.WnParam.click) then
						if (nTable.widget.WnParam.click:lower()=="left" and button=="LeftButton") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
						elseif (nTable.widget.WnParam.click:lower()=="right" and button=="RightButton") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
						elseif (nTable.widget.WnParam.click:lower()=="any") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
						end
					end
					if (nTable.widget.GetText) then
						-- Walk through all links if any
						local text=nTable.widget:GetText();
						local menu=DMMenuCreate();
						for link,atom in text:gmatch("(|c........|H(.-:.-)|h.-|h|r)") do
							menu:Add(link,function() ren.Widget_OnHyperlinkClicked(nTable.widget,atom,link,button); end);
						end
						menu:Show();		-- will not show if empty
						return;
					end
				end
			end
		end
	end
end

--DM.Color.White="|cFFFFFFFF",
--DM.Color.Green="|cFF00FF00",
--DM.Color.Red="|cFFFF0000",
--DM.Color.lBlue="|cFF6060FF",
--DM.Color.hBlue="|cFEA0A0FF",
--DM.Color.Yellow="|cFFFFFF00",

function ren:InitWidget(widget)
	widget:Hide();					-- Hide all
	if (_G[widget:GetName().."NormalTexture"]) then _G[widget:GetName().."NormalTexture"]:SetTexture(""); end
	if (_G[widget:GetName().."PushedTexture"]) then _G[widget:GetName().."PushedTexture"]:SetTexture(""); end
	if (_G[widget:GetName().."HighlightTexture"]) then _G[widget:GetName().."HighlightTexture"]:SetTexture(""); end
	if (widget.SetBackdrop) then widget:SetBackdrop(nil);end
	if (widget.SetScript) then
		widget:SetScript("OnEnter",nil); widget:SetScript("OnLeave",nil);
		if (widget.GetCursorPosition) then
			widget:SetScript("OnCursorChanged",nil);
		end
	end
	if (widget.SetChecked) then widget:SetChecked(nil); end
	if (widget._Extended) then widget._Extended=0; end
--	if (widget.SetText) then widget:SetText(""); end
--	if (widget.DMClear) then widget:DMClear(); end
	widget:ClearAllPoints();
	widget:SetHeight(0);
	widget:SetWidth(0);
end

ren.ScrMsgFrame={};
function ren.ScrMsgFrame.SetText(self,text,r,g,b)
	text=ren:toASCII(text)
	self:Clear();
	self:AddMessage(text,r,g,b);
end

ren.Bar={};
function ren.Bar._SetBarData(self,data)
	if (not self.WnParam) then return; end
	data=tonumber(data);
	if (not self.WnParam.dir) then self.WnParam.dir="right"; end
	if (not self.WnParam.max) then self.WnParam.max=100; end
	if (not self.WnParam.fill) then self.WnParam.fill="#ffff"; end
	if (not self.WnParam.empty) then self.WnParam.empty="#0000"; end
	self.WnParam.dir=self.WnParam.dir:lower();
	self.WnParam.max=tonumber(self.WnParam.max);
	local A=_G[self:GetName().."A"];
	local B=_G[self:GetName().."B"];
	local TXT=_G[self:GetName().."TXT"];
	local r,g,b,a;
	r,g,b,a=ren:SplitColor(self.WnParam.fill);		-- try to make an RGBA
	if (not r) then A:SetTexture(self.WnParam.fill); else A:SetTexture(r,g,b,a); end
	r,g,b,a=ren:SplitColor(self.WnParam.empty);		-- try to make an RGBA
	if (not r) then B:SetTexture(self.WnParam.empty); else B:SetTexture(r,g,b,a); end

	A:ClearAllPoints();
	B:ClearAllPoints();
	TXT:ClearAllPoints();
	-- => shrink/stretch alignment
	if (self.WnParam.dir=="left") then
		A:SetPoint("TOPRIGHT",self); A:SetPoint("BOTTOMRIGHT",self);
		B:SetPoint("TOPRIGHT",A,"TOPLEFT"); B:SetPoint("BOTTOMLEFT",self);
		A:SetWidth(((data/self.WnParam.max)*(self.WnParam.width-1))+1);
		TXT:SetPoint("TOPLEFT",B); TXT:SetPoint("BOTTOMRIGHT",A);
	elseif (self.WnParam.dir=="up") then
		A:SetPoint("BOTTOMLEFT",self); A:SetPoint("BOTTOMRIGHT",self);
		B:SetPoint("BOTTOMLEFT",A,"TOPLEFT"); B:SetPoint("TOPRIGHT",self);
		A:SetHeight(((data/self.WnParam.max)*(self.WnParam.height-1))+1);
		TXT:SetPoint("TOPLEFT",B); TXT:SetPoint("BOTTOMRIGHT",A);
	elseif (self.WnParam.dir=="down") then
		A:SetPoint("TOPLEFT",self); A:SetPoint("TOPRIGHT",self);
		B:SetPoint("TOPLEFT",A,"BOTTOMLEFT"); B:SetPoint("BOTTOMRIGHT",self);
		A:SetHeight(((data/self.WnParam.max)*(self.WnParam.height-1))+1);
		TXT:SetPoint("TOPLEFT",A); TXT:SetPoint("BOTTOMRIGHT",B);
	else
		A:SetPoint("TOPLEFT",self); A:SetPoint("BOTTOMLEFT",self);
		B:SetPoint("TOPLEFT",A,"TOPRIGHT"); B:SetPoint("BOTTOMRIGHT",self);
		A:SetWidth(((data/self.WnParam.max)*(self.WnParam.width-1))+1);
		TXT:SetPoint("TOPLEFT",A); TXT:SetPoint("BOTTOMRIGHT",B);
	end
end

ren.FontString={};
function ren.FontString._SetText(self,text)
	text=ren:toASCII(text)
	if (text:find("|C",1,true)) then
		local pos=text:find("|C",1,true);
		text=text.."|C";			-- add one to the end for capture
		local newtext="";			-- the new text
		local color,chunk=text:match("|C(........)(.-)|C");		-- capture capital color key
		repeat
			if (chunk:find("|r",1,true)) then
				chunk=chunk:gsub("|r","|r|c"..color);
			end
			newtext=newtext.."|c"..color..chunk;		-- rebuild
			text=text:gsub("|C(........)(.-)|C","|c"..color..chunk.."|C",1);	-- just one
			color,chunk=text:match("|C(........)(.-)|C");		-- capture capital color key
		until (not color or not chunk);
		self:SetText(text:sub(1,pos-1)..newtext);
	else
		self:SetText(text);
	end
end

ren.CheckBox={};
function ren.CheckBox._SetText(self,text)
	text=ren:toASCII(text)
	_G[self:GetName().."Text"]:SetText(text);
end

ren.SimpleHTML={};
function ren.SimpleHTML._SetText(self,text)
	text=ren:toASCII(text)
	self.DMRUTextContents=text;
	self:SetText(text);
end
function ren.SimpleHTML._GetText(self)
	return self.DMRUTextContents;
end
function ren.SetListData(self,entry)
	self:DMClear();				-- kill it
	self.defaultSort=nil;		-- keep original order
	self:DMAdd(entry);
end
function ren._SetTexture(widget,...)
	if (not _G[widget:GetName().."IconTexture"]) then return; end
	SetItemButtonTexture(widget,...);	-- probably wont support RGBA
end

function ren.SetCanvasFocus(self,id)
	if (not self.WnRenderData) then return; end
	if (not self.WnRenderData.Focus) then return; end
	if (not self.WnRenderData.Focus[id]) then return; end
	if (self:GetParent().SetVerticalScroll) then
		self:GetParent():SetVerticalScroll(self.WnRenderData.Focus[id]);
	end
end

--	DM:Chat("Clicked: "..linkdata);	-- essential center - "H" part
--	DM:Chat("Clicked: "..link);		-- Complete link text
--	DM:Chat("Clicked: "..button);	-- "LeftButton"
function ren.Widget_OnHyperlinkClicked(widget,linkdata,link,button)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_CLICKED",linkdata,link,button);
end
function ren.Widget_OnHyperlinkEnter(widget,linkdata,link)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_ENTER",linkdata,link);
end
function ren.Widget_OnHyperlinkLeave(widget,linkdata,link)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_LEAVE",linkdata,link);
end

function ren.SendEvent(widget,event,arg1,arg2,arg3,arg4)
	local frame=widget:GetParent():GetParent():GetParent();	-- canvas,scrollframe,appwindow
	local handler=frame:GetScript("OnEvent");
	if (not handler) then
		frame=widget:GetParent():GetParent();
		handler=frame:GetScript("OnEvent");
		if (not handler) then
			frame=widget:GetParent();
			handler=frame:GetScript("OnEvent");
			if (not handler) then DEFAULT_CHAT_FRAME:AddMessage("No handler found for a canvas"); return; end
		end
	end
--	frame:OnEvent(event,canvas,arg1,arg2,arg3,arg4)
	handler(frame,event,widget:GetParent(),arg1,arg2,arg3,arg4);	-- simulated self
end

function ren.Widget_OnEnter(widget)
	if (widget.WnParam.tip) then
		ren.SendEvent(widget,"DMEVENT_WIDGET_ENTER",widget.WnParam.tip,widget.WnContext,widget.WnParam);
	end
end

function ren.Widget_OnLeave(widget)
	if (widget.WnParam.tip) then
		ren.SendEvent(widget,"DMEVENT_WIDGET_LEAVE");
	end
end

--function ren.Widget_OnCursorChanged(widget,x,y,width,height)
function ren.Widget_OnCursorChanged(widget)
	if (widget.WnParam.onchange) then
--print(widget.WnParam.onchange,widget:GetText());
		DM:StoreByText(widget.WnParam.onchange,widget,widget:GetText());
	end
end

function ren.GetContextData(canvas,key,val)
	local data={};
	for _,wg in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for _,wData in pairs(wg) do					-- All widgets within group
			if (wData.Used) then
				local element=wData.widget.WnElement;
				local GetMyData=canvas.WnRenderData.Meta["<elements>"][element].IF.GetData;
				if (GetMyData) then
					if (wData.Used and wData.widget.WnContext[key] and wData.widget.WnContext[key]==val) then
						table.insert(data,{
							param=DM:CopyTable(wData.widget.WnParam),
							context=DM:CopyTable(wData.widget.WnContext),
							data=wData.widget[GetMyData](wData.widget),
							element=element,
							widget=wData.widget,
						});
					end
				end
			end
		end
	end
	return data;
end

function ren.GetObject(canvas,parameter,value)
	for _,wg in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for _,wData in pairs(wg) do					-- All widgets within group
			if (wData.Used and wData.widget.WnParam and wData.widget.WnParam[parameter]==value) then
				return wData.widget;
			end
		end
	end
end

function ren.PageButtonClicked(button,mbutton)
--print("button",mbutton);
	if (not button) then DM:Chat("No button provided"); return; end
	local canvas=button.WnCanvas;

	-- run base radiobutton functionality
	if (button.WnContext.group) then
		local valid=nil;
		local form=canvas:GetForm("group",button.WnContext.group);
		for _,d in pairs(form) do
			if (d.widget==button and d.element=="<radio>") then valid=true; break; end
		end
		if (valid) then
			for _,d in pairs(form) do
				if (d.element=="<radio>") then d.widget:SetChecked(nil); end
			end
			button:SetChecked(true);
		end
	end

--DM:Chat("Button clicked: "..button:GetName());
	if (button.WnParam.click) then						-- param "click" supplied
--		DM:Chat("An element click has been received");
		if (button.WnParam.click:lower()=="left" and mbutton~="LeftButton") then return; end
		if (button.WnParam.click:lower()=="right" and mbutton~="RightButton") then return; end
--		ren.SendEvent(button,"DMEVENT_BUTTON_CLICKED",button.WnContext,button.WnParam);
--		return;
	end
	ren.SendEvent(button,"DMEVENT_BUTTON_CLICKED",button.WnContext,button.WnParam,mbutton);
	if (not button.WnParam) then						-- No params in XML
		DM:Chat("No action associated with this button.");
	elseif (button.WnParam.send) then						-- param "send" supplied
		if (button.WnContext[button.WnParam.send]) then		-- It's send-data exists in the context
			local val=button.WnContext[button.WnParam.send];
			local transData={};
			local stored=nil;
			for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
				local GetMyData=canvas.WnRenderData.Meta["<elements>"][button.WnElement].IF.GetData;
				for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
					if (canvas.WnRenderData.Meta["<elements>"][wwTable.widget.WnElement].IF.GetData) then
						if (wwTable.Used and wwTable.widget.WnParam.id and wwTable.widget.WnContext[button.WnParam.send]==val) then
							transData[wwTable.widget.WnParam.id]=wwTable.widget[GetMyData](wwTable.widget);
							stored=true;
						end
					end
				end
			end
			-- We've got data, so send it
			if (stored) then
--					DM.Net:ConnectW(prefix,cbRX,receiver);
				local canvas=button:GetParent();
				if (not canvas.SECTIONDATA) then return; end			-- No meta-data
				if (not canvas.SECTIONDATA.Meta) then return; end			-- No meta-data
				if (not canvas.SECTIONDATA.Meta.Server) then return; end		-- No server to send to
				-- Make a queue for polling and sending depending on result
				DM.WoWnet.Session:SendPublic("ping:"..canvas.SECTIONDATA.Meta.Server);	-- Do ping
				if (not DM.WoWnet.Session.ResolveWait) then				-- Nothing else waiting
					DM.WoWnet.Session.ResolveWait={
						Table=transData,					-- Actual data to send
						Server=canvas.SECTIONDATA.Meta.Server,	-- Server to send to
						PipeToon=nil;						-- Who's the toon
					};
				end
			end
		end
	else
--		DM:Chat("No known action associated with this button.");
	end
end


-- The following code enables context-code, like so:
-- <set form="guild-application"/>  -- Starting form for guild application (multiple forms allowed)
-- <edit req="1"/>				  -- Assigned to form "guild-application" (requires at least one character data)
-- <button send="form"/>			-- Assigned to form "guild-application"
-- <set form="guild-log-in"/>	   -- Starting form for log-in, guild application form is done
-- <edit ID="LIname"/>
-- <edit ID="LIpass"/>
-- <set form=""/>					-- No form active
-- Following this, pressing the button will collect data from all active elements
-- associated with the same context. In this case "form".
function ren:SetContext(canvas,param)
	if (not param) then param={}; end
	canvas.WnRenderData.WnContext=ren:Mix(param,canvas.WnRenderData.WnContext);
	for entry,eVal in pairs(canvas.WnRenderData.WnContext) do
		if (eVal=="") then canvas.WnRenderData.WnContext[entry]=nil; end		-- Remove emptied parameters
	end
end

function ren:Mix(paramA,paramB)
	if (not paramA) then paramA={}; end
	local param=DM:CopyTable(paramB);		-- Make copy
	return DM:CopyTable(paramA,param);		-- Merge and return
end

function ren:CopyTextAttributes(wF,wT)
	wT:SetFont(wF:GetFont());
	wT:SetSpacing(wF:GetSpacing());
end

function ren:toHTML(s)
	s=s or "";
	s=s:gsub("\"","&quot;");
	s=s:gsub("<","&lt;");
	s=s:gsub(">","&gt;");
	return s;
end

function ren:toASCII(s)
	s=s or "";
	s=s:gsub("&quot;","\"");
	s=s:gsub("&lt;","<");
	s=s:gsub("&gt;",">");
	return s;
end

-- Handle "glue" and "flow"
function ren:SetPosition(canvas,widget,param,swap,onlysize)
	if (not param) then param={}; end
	local glue,flow,align,pad=param.glue,param.flow,param.align,param.pad;		-- Make copies
	local bottomtype="stack";

	if (swap) then
--		ren:CopyTextAttributes(widget,canvas.MeasureFontString)
		swap=widget;
		widget=canvas.MeasureFontString;
	end

	widget:ClearAllPoints();
	local cWidth=canvas:GetWidth();
	local hSet,wSet;
	if (param.width) then
		local value=tostring(param.width);
		if (value:sub(-1)=="%") then
			value=tonumber(value:sub(1,-2));			-- Get the percentage
			value=(cWidth/100)*value;	-- Convert to points
		end
--DM:Chat("Width: "..tonumber(value));
		if (widget._Extended) then
			widget._Extended=value-param.height;
			value=param.height;
		end
		widget:SetWidth(tonumber(value));				-- Set provided value directly
		wSet=true;
	end
	if (param.height) then
		widget:SetHeight(tonumber(param.height));		-- Set provided value directly
		hSet=true;
	end
	if (onlysize) then return; end

	-- Check provided information
	if (not glue and not flow and not align) then align="LEFT"; end
	if (not align) then align=""; end
	if (not glue) then glue=""; end
	if (not flow) then flow=""; end
	if (not pad) then pad=0; end
	align=align:upper();
	glue=glue:upper();
	flow=flow:upper();

--DM:Chat("g:"..glue.." f:"..flow.." a:"..align);

	-- Set text adjustment. "align" has presedence as it does that for positioning as well.
	local adjust;
	adjust="LEFT";		-- Since we are left-to-right readers
	if (flow=="LEFT") then adjust="RIGHT"; end
	if (flow=="DOWN") then adjust="CENTER"; end
	if (align=="LEFT" or align=="RIGHT" or align=="CENTER") then adjust=align; end

	local lastE=0; if (canvas.WnRenderData.LastWidget) then lastE=(canvas.WnRenderData.LastWidget._Extended or 0); end
	local lockside,bottomadd="LEFT",pad;
	-- Align with canvas in stead of widget
	if (align=="LEFT" or (not canvas.WnRenderData.LastWidget and glue=="LEFT")) then
		widget:SetPoint("TOPLEFT",canvas,"TOPLEFT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		canvas.WnRenderData.LockH=0;
		adjust="LEFT";
	elseif (align=="RIGHT" or (not canvas.WnRenderData.LastWidget and glue=="RIGHT")) then
		widget:SetPoint("TOPRIGHT",canvas,"TOPRIGHT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="RIGHT";
		canvas.WnRenderData.LockH=cWidth-0;
		adjust="RIGHT";
	elseif (align=="CENTER" or (not canvas.WnRenderData.LastWidget and glue=="TOP")) then
		widget:SetPoint("TOP",canvas,"TOP",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="CENTER";
		canvas.WnRenderData.LockH=cWidth/2;
		adjust="CENTER";
	elseif (glue=="LEFT") then
		if (flow=="DOWN") then widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"TOPRIGHT",pad+lastE,0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMLEFT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",pad+lastE,0); bottomtype="BOTTOMUP";
		else widget:SetPoint("LEFT",canvas.WnRenderData.LastWidget,"RIGHT",pad+lastE,0); bottomtype="CENTERSIDE"; end
		adjust="LEFT";
		bottomadd=0;
	elseif (glue=="RIGHT") then
		if (flow=="DOWN") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"TOPLEFT",-(pad+lastE),0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",-(pad+lastE),0); bottomtype="BOTTOMUP";
		else widget:SetPoint("RIGHT",canvas.WnRenderData.LastWidget,"LEFT",-(pad+lastE),0); bottomtype="CENTERSIDE"; end
		lockside="RIGHT";
		adjust="RIGHT";
		bottomadd=0;
	elseif (glue=="TOP") then
		if (flow=="DOWN") then widget:SetPoint("TOP",canvas.WnRenderData.LastWidget,"BOTTOM",0,0-(pad+lastE)); lockside="CENTER"; adjust="CENTER";
		elseif (flow=="LEFT") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",0,0-(pad+lastE)); lockside="RIGHT"; adjust="RIGHT";
		else widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",0,0-(pad+lastE)); lockside="LEFT"; adjust="LEFT"; end
	end
	if (widget.SetJustifyV) then widget:SetJustifyV("TOP"); end
	if (widget.SetJustifyH) then widget:SetJustifyH(adjust); end

	-- Expand widget if text is bigger than said area
	if (not wSet and widget.GetStringWidth) then
		local sWidth=widget:GetStringWidth();

		if (lockside=="LEFT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+pad;
			if (canvas.WnRenderData.LockH+sWidth>cWidth) then
				widget:SetWidth(cWidth-canvas.WnRenderData.LockH);
			end
--			widget:SetPoint("RIGHT");
		elseif (lockside=="RIGHT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-pad;
			if (canvas.WnRenderData.LockH-(sWidth)<0) then
				widget:SetWidth(canvas.WnRenderData.LockH);
			end
--			widget:SetPoint("LEFT");
		elseif (lockside=="CENTER") then
			if (canvas.WnRenderData.LockH-(sWidth/2)<0) then
				widget:SetWidth(cWidth);
			end
--			widget:SetPoint("LEFT");
--			widget:SetPoint("RIGHT");
		end
	end


	if (not hSet and widget.GetStringHeight) then
		widget:SetHeight(widget:GetStringHeight());
	end

	-- Set new lock-side
	if (lockside=="LEFT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+widget:GetWidth();
	elseif (lockside=="RIGHT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-widget:GetWidth();
	elseif (lockside=="CENTER") then
		-- Assume continuous position, even though that may be wrong when the center is center of an element
	end

	-- Set the visually lowest point of the rendering
	canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+bottomadd;		-- Add any vertical padding
	if (bottomtype=="stack") then
		canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+widget:GetHeight();
	elseif (bottomtype=="TOPDOWN") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight());
		end
	elseif (bottomtype=="CENTERSIDE") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight())/2;
		end
	end

	if (swap) then
--DM:Chat("s:"..math.floor(widget:GetHeight())..","..math.floor(widget:GetWidth()));
--		widget:Show();
		swap:ClearAllPoints();
		swap:SetParent(widget:GetParent());
		swap:SetWidth(widget:GetWidth()*1.0065+0.5);
--		if (wSet) then swap:SetWidth(widget:GetWidth()); end
		swap:SetHeight(widget:GetHeight()+0.5);
		local index=1;
		while(index<=widget:GetNumPoints()) do swap:SetPoint(widget:GetPoint(index)); index=index+1; end
--		widget:Hide();
		widget=swap;
	end
--DM:Chat("w:"..math.floor(widget:GetHeight())..","..math.floor(widget:GetWidth()));
end

-- Find an unused widget or create a new
function ren:GetWidget(canvas,element,widget,param)
	if (not param) then param={}; end

	if (element=="<img>" and param.tip) then
		element="<img_i>";
		widget=canvas.WnRenderData.Meta["<elements>"][element];
	end

	local typename=widget.widget;
	local inheritsys=widget.inheritsys;
	if (param.inheritsys) then inheritsys=param.inheritsys; end		-- Note the lower case
	if (not canvas.WnRenderData.Widget[element]) then canvas.WnRenderData.Widget[element]={}; end
	local TheName=nil;
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget[element]) do
		if (not wTable.Used and not canvas.WnRenderData.Widget[element][wEntry].widget.SPECIALNOREUSE) then
			if (canvas.WnRenderData.Widget[element][wEntry].inheritsys==inheritsys) then
				TheName=wEntry;
				break;
			end
		end
	end

	-- Create new if needed
	local e=nil;
	if (not TheName) then
		TheName="WoWnet-widget-"..DM:Random(0,1000000);		-- Fractional million
		canvas.WnRenderData.Widget[element][TheName]={};
		e=canvas.WnRenderData.Widget[element][TheName];
		if (self:Flag(canvas,widget,"HaveContainer")) then
			local creator="Create"..typename;
			e.widget=canvas[creator](canvas,TheName);	-- Simulate colon
		else
			e.widget=CreateFrame(typename,TheName,canvas,inheritsys);
			e.inheritsys=inheritsys;
			if (self:Flag(canvas,widget,"Extended")) then
				e.widget._Extended=0;
			end
		end
		if (self:Flag(canvas,widget,"List")) then
			e.widget._SetText=ren.SetListData;
		elseif (self:Flag(canvas,widget,"SimpleHTML")) then
			e.widget._SetText=ren.SimpleHTML._SetText;
			e.widget._GetText=ren.SimpleHTML._GetText;
		elseif (self:Flag(canvas,widget,"DeadEditBox")) then
			e.widget:SetHyperlinksEnabled(true);
			e.widget:SetTextInsets(-1,-2,-1,0);	-- lrtb
--			e.widget:SetTextInsets(0,0,0,0);	-- lrtb
			e.widget:SetMultiLine(true);
			e.widget:EnableKeyboard(nil);
			e.widget:SetAutoFocus(nil);
			e.widget:SetScript("OnHyperlinkClick",ren.Widget_OnHyperlinkClicked);
			e.widget:SetScript("OnHyperlinkEnter",ren.Widget_OnHyperlinkEnter);
			e.widget:SetScript("OnHyperlinkLeave",ren.Widget_OnHyperlinkLeave);
		elseif (self:Flag(canvas,widget,"ScrMsgFrame")) then
			e.widget.SetText=ren.ScrMsgFrame.SetText;
			e.widget:SetHyperlinksEnabled(true);
			e.widget:SetFading(nil);
			e.widget:SetIndentedWordWrap(nil);
			e.widget:SetMaxLines(10000);
			e.widget:SetResizable(nil);	-- user resizable off
		elseif (self:Flag(canvas,widget,"CheckBox")) then
			e.widget._SetText=ren.CheckBox._SetText;
		elseif (self:Flag(canvas,widget,"Bar")) then
			e.widget._SetBarData=ren.Bar._SetBarData;
			e.widget:CreateTexture(TheName.."A");
			e.widget:CreateTexture(TheName.."B");
			e.widget:CreateFontString(TheName.."TXT");
--		elseif (self:Flag(canvas,widget,"FontString")) then
		elseif (self:Flag(canvas,widget,"FontInstance")) then
			e.widget._SetText=ren.FontString._SetText;
		end
	else
		e=canvas.WnRenderData.Widget[element][TheName];
		e.widget:SetParent(canvas);			-- In case it was moved earlier
		wipe(e.widget.WnContext);				-- Clear the context for the element
		e.widget.WnContext=nil;				-- And throw it away
		if (e.widget.WnParam) then
			wipe(e.widget.WnParam);				-- Clear the parameters for the element
			e.widget.WnParam=nil;					-- And throw it away
		end
	end
	-- Set up the widget a bit
	e.widget.WnElement=element;
	if (e.widget.SetDrawLayer) then
		e.widget:SetDrawLayer("ARTWORK");
	end
	e.widget.WnContext=DM:CopyTable(canvas.WnRenderData.WnContext);	-- Insert current context in the element
	e.Used=true;
	e.widget.WnParam=DM:CopyTable(param);
	e.widget.WnCanvas=canvas;
	-- Make fake texture handler for those who can't
	if (not e.widget.SetTexture) then
		e.widget.SetTexture=ren._SetTexture;
	end
	-- Set some font for widget with fontstrings
	if (self:Flag(canvas,widget,"FontString")) then
		e.widget:SetFontObject("GameFontNormal");
	end

	ren:InitWidget(e.widget);

	-- Set mouse-over handler for those who can
	if (e.widget.SetScript) then
		e.widget:SetScript("OnEnter",ren.Widget_OnEnter);
		e.widget:SetScript("OnLeave",ren.Widget_OnLeave);
		if (e.widget.GetCursorPosition) then
			e.widget:SetScript("OnCursorChanged",ren.Widget_OnCursorChanged);
		end
	end
	-- Set handlers
	if (self:Flag(canvas,widget,"Button")) then
		-- check for button that must not have a click handler
		if (self:Flag(canvas,widget,"NoAutoEvent") and (param and not param.click)) then
			e.widget:SetScript("OnClick",nil);
		else
			e.widget:SetScript("OnClick",ren.PageButtonClicked);
			if (e.widget.RegisterForClicks) then e.widget:RegisterForClicks("AnyUp"); end
		end
	end

	return e.widget;
end

-- This function takes a parameter table, and assumes that it inherits
-- all the way back to default.
function ren:SetFont(canvas,widget,param)
	local file,size,flags,r,g,b,a;
	if (not param) then param={}; end
	if (not widget.SetFont) then
	end

	-- Start with default settings
	file=canvas.WnRenderData.Meta.font.Default.file;
	size=canvas.WnRenderData.Meta.font.Default.size;
	flags=canvas.WnRenderData.Meta.font.Default.flags;
	r,g,b,a=self:SplitColor(canvas.WnRenderData.Meta.font.Default.color);

	if (param.font and canvas.WnRenderData.Meta.font[param.font]) then			-- Font supplied and it exists
		file,size,flags,r,g,b,a=self:MakeFont(widget,file,size,flags,canvas.WnRenderData.Meta.font[param.font],canvas.WnRenderData.Meta,r,g,b,a);
	end
--print(file,size,flags);
	widget:SetFont(file,tonumber(size),flags);
	widget:SetTextColor(r,g,b,a);					-- set it
--print(r,g,b,a);
	widget:SetShadowOffset(0,0);
end

function ren:MakeFont(widget,file,size,flags,font,meta,r,g,b,a)
	if (not font) then return file,size,flags; end

	if (font.inherits and meta.font[font.inherits]) then	-- Inherits from some other font
		file,size,flags,r,g,b,a=ren:MakeFont(widget,file,size,flags,meta.font[font.inherits],meta,r,g,b,a);
	end
	if (font.file) then file=font.file; end
	if (font.size) then size=font.size; end
	if (font.flags) then flags=font.flags; end
	if (font.color) then r,g,b,a=self:SplitColor(font.color); end
	return file,size,flags,r,g,b,a;
end

function ren:SplitColor(color)
	local r,g,b,a;
	if (not color) then return; end
	if (color:find("#",1,true)==1) then
		color=color:sub(2);
		if (color:len()==8) then		-- Standard full-resolution mode 0-255
			a=tonumber("0x"..color:sub(1,2)); if (a>0) then a=a/255; end
			r=tonumber("0x"..color:sub(3,4)); if (r>0) then r=r/255; end
			g=tonumber("0x"..color:sub(5,6)); if (g>0) then g=g/255; end
			b=tonumber("0x"..color:sub(7,8)); if (b>0) then b=b/255; end
		elseif (color:len()==4) then	-- Simple html-mode 0-15
			a=tonumber("0x"..color:sub(1,1)); if (a>0) then a=a/15; end
			r=tonumber("0x"..color:sub(2,2)); if (r>0) then r=r/15; end
			g=tonumber("0x"..color:sub(3,3)); if (g>0) then g=g/15; end
			b=tonumber("0x"..color:sub(4,4)); if (b>0) then b=b/15; end
		end
	end
	return r,g,b,a;
end

function ren:SetInterface(canvas,element,class)
	if (not canvas.WnRenderData.Meta["<elements>"][element]) then return; end
--DM:Chat("IF: "..element);
	canvas.WnRenderData.Meta["<elements>"][element].Class=class;
	if (not canvas.WnRenderData.Meta["<elements>"][element].IF) then canvas.WnRenderData.Meta["<elements>"][element].IF={}; end
	-- Check for implied interfaces
	if (not bit.band(class,ren.Class.UIObject)) then DM:Chat("<"..element.."> does not have UIObject",1); return; end
	if (not bit.band(class,ren.Class.Region)) then DM:Chat("<"..element.."> does not have Region",1); return; end
	-- Clear the common programming interface
	canvas.WnRenderData.Meta["<elements>"][element].IF.SetData=nil;		-- No method
	canvas.WnRenderData.Meta["<elements>"][element].IF.GetData=nil;		-- No method

	-- Set the common programming interface
	if (bit.band(class,ren.Class.Bar)==ren.Class.Bar) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetBarData";
	elseif (bit.band(class,ren.Class.CheckBox)==ren.Class.CheckBox) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetChecked";
	elseif (bit.band(class,ren.Class.SimpleHTML)==ren.Class.SimpleHTML) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";	-- These are hooked
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="_GetText";
	elseif (bit.band(class,ren.Class.Texture)==ren.Class.Texture) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetTexture";
	elseif (bit.band(class,ren.Class.Button)==ren.Class.Button) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,ren.Class.ScrMsgFrame)==ren.Class.ScrMsgFrame) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
--	elseif (bit.band(class,ren.Class.FontString)==ren.Class.FontString) then
--		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
--		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
--	elseif (bit.band(class,ren.Class.FontInstance)==ren.Class.FontInstance) then
--		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
--		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,ren.Class.FontInstance)==ren.Class.FontInstance) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,ren.Class.List)==ren.Class.List) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
	end
end

function ren:Flag(canvas,widget,flag)
	if (bit.band(widget.Class,ren.Class[flag])==ren.Class[flag]) then
		return true;
	end
	return nil;
end


function ren:PullParameters(text)
--DM:Chat("P:"..text);
	local param={};
	local start=0;
	text=strtrim(text);
	repeat
		start=start+1;
		local pE=text:find("=\"",start,true)-1;
		local dE=text:find("\"",pE+3,true)-1;
		param[text:sub(start,pE):lower()]=text:sub(pE+3,dE);
--DM:Chat("P->:"..text:sub(start,pE):lower().."="..text:sub(pE+3,dE));
		start=dE+2;									-- Position at space between parameters
	until (text:sub(start,start)~=" ");
	return param;
end



-- * The section must start at the first character
-- * Does NOT support nested same-type top-tags
-- Returns:
--   Section name: String. Contents of first tag without parameters
--	 Parameters: Table. All parameters where entry is param-name and value is string value
--		   Data: The entire contents between starting and ending tags
--	  Remaining: The remainder of the data-block
function ren:PullSection(data)
--DM:Chat("PullSection...",1,0,1);
	local locS=data:find(" ",1,true);
	local locE=data:find(">",1,true);

	-- locTagStart is by definition always 1
	-- locTagNameStart is by definition always 2
	local locTagNameEnd,locTagEnd
	local params=nil;

	if (locS and locS<locE) then	-- With parameters
		locTagNameEnd=locS-1;
		locTagEnd=locE;
		if (data:find("/>",locTagEnd-1,true)==locTagEnd-1) then		-- No data included
			params=self:PullParameters(data:sub(locS+1,locTagEnd-2));
			-- Return the distilled data (tag,params,data,remaining)
--DM:Chat("R:"..strtrim(data:sub(locTagEnd+1,locTagEnd+1+25)));
			return data:sub(1,locTagNameEnd)..">",params,nil,strtrim(data:sub(locTagEnd+1));
		end
		params=self:PullParameters(data:sub(locS+1,locTagEnd-1));
	else 							-- Without parameters
		locTagNameEnd=locE-1;
		locTagEnd=locE;
	end
	local locEndTagStart,locEndTagEnd=data:find("</"..data:sub(2,locTagNameEnd)..">",locE,true);
--DM:Chat("D:"..strtrim(data:sub(locTagEnd+1,locEndTagStart-1)));
--DM:Chat("R:"..strtrim(data:sub(locEndTagEnd+1,locEndTagEnd+1+25)));
	return data:sub(1,locTagNameEnd)..">",params,strtrim(data:sub(locTagEnd+1,locEndTagStart-1)),strtrim(data:sub(locEndTagEnd+1));
end

function ren:PullParameter(data)
	data=strtrim(data);
	local loc=data:find("=\"",1,true);
	if (not loc) then return nil,nil,""; end
	local param=data:sub(1,loc-1);
	data=data:sub(loc+2);
	loc=data:find("\"",1,true);
	local value=data:sub(1,loc-1);
	data=data:sub(loc+1);
	data=strtrim(data);
	return param:lower(),value,data;
end
