--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2014  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=2.9991;

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end
local DM=DuckMod[TV];
local _;

local DUCKNET_HEARTBEAT = 1/5;			-- 200ms ticker (5 fps)


DM.WoWnet={
	Socket={			--> Client Sockets
		List={},		-- existing sockets
		API={},			-- socket api
	},
	Listen={			--> Server Sockets
		List={},		-- existing sockets
		API={},			-- socket api
	},
	Public={			--> Public Sockets
		List={},		-- existing sockets
		API={},			-- socket api
		Sent={},		-- sent public messages
		Incoming={},	-- incoming public raw text
	},
	Generic={			--> Generic Sockets
	},
	MT={},				--> Multitasking
	Passed=0,			-- time passed
};
local wn=DM.WoWnet;
local sct=wn.Socket;
local sapi=sct.API;
local lis=wn.Listen;
local lapi=lis.API;
local pub=wn.Public;
local papi=pub.API;
local gen=wn.Generic;

local WN_ID="wn";

local wnBoot=time()+10;


local function dbg(h,t)
	print(h,":");
	for _i,_d in pairs(t) do print(_i,_d); end
end


-- Set up for handling
function wn:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-WoWnet-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod WoWnet UI frame",1); return; end
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-WoWnetWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod WoWnet WM frame",1); return; end
	end
	self.TheFrame:RegisterEvent("CHAT_MSG_CHANNEL");			-- Public WoWnet
	self.TheFrame:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");		-- Public WoWnet info
	self.TheFrame:SetScript("OnEvent",self.OnEvent);
	self.TheFrame:SetScript("OnUpdate",self.Tick);
	self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
end

function wn:Tick(elapsed)
--	table.insert(pub.Incoming,{ ad=add, pr=prot, to=arg2, da=data, ep=ep } );
	wn.Passed=wn.Passed+elapsed;
	if (wn.Passed<DUCKNET_HEARTBEAT) then return; end
	wn.Passed=0;				-- inaccurate to account for time-snags
	if (not pub.Incoming[1]) then return; end

	-- handle public input
	local entry=table.remove(pub.Incoming,1);
	local buf=pub:PublicRX(entry.da,entry.to);			-- in case of broadcast messages
	if (buf) then										-- handle them
		-- check if a generic socket is not busy (IsIdle)
		local sck;
		for _,_s in pairs(gen) do
			if (_s:IsIdle()) then sck=_s; break; end					-- find one idle
		end
		-- - if not, make one more
		if (not sck) then
			sck=DM.Com2:CreateInterface(WN_ID,nil,nil,nil,nil,"WHISPER");	-- create new
			table.insert(gen,sck);										-- add to list
		end
		-- make it ready
		sck.ep=entry.ep;
		sck:SetReceiver(entry.to);
		-- send buf to selected static socket
		sck:Send(buf,entry.ep);
		sck.ep=0;			-- keep it generic
		return;
	end

	-- listening interfaces
	for _,_i in pairs(lis.List) do
		if (add==_i.address and prot==_i.protocol and _i.hash==hash) then
			_i.fIn(nil,{a=entry.ad,p=entry.pr,t=entry.to,d=entry.da},entry.ep,nil,entry.to,nil);
		end
	end
end


-- <address>:<protocol>:<ep>:<protocol specific contents>
function pub:SplitInput(text)
	return text:match("^wn:(.-):(.-):(.-):(.+)$");
end

-- Other
function wn:OpenSocket(add,prot,toon,cbRX)
	local socket={};
	socket.address=add;
	socket.protocol=prot;
	socket.ep=DM:Hash(add..prot);
	socket.Close=sapi.Close;
	socket.Send=sapi.Send;
	socket.Go=sapi.Go;
	socket.fIn=function(_,data,ep,_,sender,_)
		if (not ep or ep~=socket.ep) then return; end
		if (not data or type(data)~="table" or data.a~=socket.address or data.p~=socket.protocol) then return; end
		if (data.t~=UnitName("player")) then return; end
		cbRX(socket,sender,data.d);
	end
	socket.interface=DM.Com2:CreateInterface(WN_ID,socket.fIn,nil,nil,nil,"WHISPER");
	socket.interface:SetReceiver(toon);
	table.insert(sct.List,socket);
	return socket;
end

function wn:Listen(add,prot,cbRX,desc)
	local listen={ Toons={}, };
	listen.address=add:lower();
	listen.protocol=prot:lower();
	listen.ep=DM:Hash(add..prot);
	listen.description=desc or "";
	listen.PO=pub.PO;

	listen.fIn=function(_,data,ep,_,sender,_)
		if (data.a~=listen.address or data.p~=listen.protocol) then return; end
		-- create transmitting interface for a single toon
		if (not listen.Toons[sender]) then							-- this toon's first time
			listen.Toons[sender]={ interface=DM.Com2:CreateInterface(WN_ID,nil,nil,nil,nil,"WHISPER"); };
			listen.Toons[sender].interface:SetReceiver(sender);
		end
		listen:cbRX(sender,data.d,
			function (senddata)			-- reply-function for servers
				listen.Toons[sender].interface:Send({a=data.a,p=data.p,t=data.t,d=senddata},ep);
			end);
	end
	-- create listening interface
	listen.interface=DM.Com2:CreateInterface(WN_ID,listen.fIn,nil,nil,nil,"WHISPER");

	table.insert(lis.List,listen);
	return listen;
end

function wn:Public(cbSR)
	local public={};
	public.address=UnitName("player"):lower()..string.format("%X",tostring(math.random(0x7FFF)));
	public.protocol="wownet";
	public.ep=DM:Hash(public.address..public.protocol);		-- End point
	public.PO=papi.PO;
	public.FindProtocol=papi.FindProtocol;
	public.FindAddress=papi.FindAddress;
	public.Search=papi.Search;

	public.fIn=function(_,data,ep,_,sender,_)	-- data straight from private socket
		if (ep and ep~=public.ep) then return; end
--		if (cbSR) then cbSR(public,data,sender); end			-- relay data to addon
		if (cbSR) then cbSR(data); end			-- relay data to addon
	end
	-- create listening interface for private replies
	public.interface=DM.Com2:CreateInterface(WN_ID,public.fIn,nil,nil,nil,"WHISPER");

	table.insert(pub.List,public);
	return public;
end


--==>> Socket API
function sapi.Close(socket)
	for i,s in pairs(sct.List) do
		if (s==socket) then sct.List.interface:Close(); table.remove(sct.List,i); return; end
	end
end

function sapi.Send(socket,toon,data)
	socket.interface:SetReceiver(toon);
	socket.interface:Send({a=socket.address,p=socket.protocol,t=toon,d=data},socket.ep);
end

function sapi.Go(socket,address,protocol,toon,location)
	if (not address or not protocol or not toon) then return; end
	local name="index";
	if (not location) then location="/"; end
	if (not location:find("/",1,true)) then		-- does not have any "/"
		name=location;							-- assume it's the document name
		location="/";							-- assume it's root
	end
	if (not location:match("/$")) then			-- does not end in "/"
		local buf={strsplit("/",location)};
		for _,d in ipairs(buf) do name=d; end	-- get last without using crappy table operator
	end
	socket.address=address;
	socket.protocol=protocol;
	socket:Send(toon,{location=location, name=name, });
end

--==>> Public API
function papi.PO(public,data)			-- public Out
	if (not wn.Connected) then return; end
	local tmp=WN_ID..":"..public.address..":"..public.protocol..":"..public.ep..":"..data;
	pub.Sent[tmp]=true;
	SendChatMessage(tmp,"CHANNEL",nil,wn.Connected);
end

function papi.FindProtocol(public,prot)
	public:PO("findprot:"..prot);
end

function papi.FindAddress(public,add)
	public:PO("findadd:"..add);
end

function papi.Search(public,search)
	public:PO("search:"..search);
end


-- the data presented here is from the public channel, so simple
-- textual WoW-allowed strings only.
-- NOTE: This channel will be affected by drunkenness.
function pub:PublicRX(data,sender)
	if (type(data)~="string") then return nil; end
	local buf={};
	data=strtrim(data);
print("here")
	-- Regular listen socket
	if (data:find("findprot:",1,true)==1) then			--==>> List sockets for this protocol
		data=data:match("^findprot:(.+)$");
		data=data:lower();
		-- check all listening interfaces
		for _,li in pairs(lis.List) do
			if (li.protocol==data) then
				table.insert(buf,{address=li.address,protocol=li.protocol,description=li.description});
			end
		end
		if (not next(buf)) then return nil; end			-- no matches
		buf.Contents="Protocol";
	elseif (data:find("findadd:",1,true)==1) then		--==>> List sockets for this address
		data=data:match("^findadd:(.+)$");
		data=data:lower();
		-- check all listening interfaces
		for _,li in pairs(lis.List) do
			if (li.address==data) then
				table.insert(buf,{address=li.address,protocol=li.protocol,description=li.description});
			end
		end
		if (not next(buf)) then return nil; end			-- no matches
		buf.Contents="Address";
	elseif (data:find("search:",1,true)==1) then		--==>> List sockets for this search
		data=data:match("^search:(.+)$");
		data=data:lower();
		-- check all listening interfaces
		for _,li in pairs(lis.List) do
			if (li.address:find(data,1,true) or li.description:lower():find(data,1,true)) then
				table.insert(buf,{address=li.address,protocol=li.protocol,description=li.description});
			end
		end
		if (not next(buf)) then return nil; end			-- no matches
		buf.Contents="Search";
	end
	buf.Entry=data;
	return { [sender]=buf, };
end


--==>> Listen API
function lapi.Close(listen)
	for i,s in pairs(lis.List) do
		if (s==listen) then
			listen.interface:Close();
			for _,_i in pairs(listen.Toons) do _i.interface:Close(); end
			table.remove(lis.List,i);
			return;
		end
	end
end


-- complete link presentation: "|cff808080|Hwownet:theadresshere|h[a thingie]|h|r"
-- only link: "wownet:theadresshere"
-- full link: wownet:address.protocol@character/location hierarchy/optional file
-- 1st: wownet:address.protocol@character/location hierarchy/optional file
-- 2nd: wownet:address.protocol/location hierarchy/optional file
-- 3rd: wownet:address/location hierarchy/optional file
function wn:ResolveLink(link)
	link=strtrim(link);
	if (not link:find("/",1,true)) then link=link.."/index"; end	-- make sure it has a location
	local address,protocol,toon,location=link:match("^wownet:(.-)%.(%w-)@(.-)(/.*)$");
	if (not address) then address,protocol,location=link:match("^wownet:(.-)%.(%w-)(/.*)$"); end
	if (not address) then address,location=link:match("^wownet:(.-)(/.*)$"); end
	if (not address or not location) then return; end		-- could not read link
	return address,protocol,toon,location;
end

function wn:FormatLink(address,protocol,toon,location,displaytext)
	if (not address) then return displaytext or ""; end
	local uselink;
	if (displaytext) then uselink=true; end
	local top="|cffA0FFA0|H";
	local bot="]|h|r";
	if (not location or location=="") then location="/"; end
	local link="wownet:"..address; if (uselink) then displaytext=link; end
	if (not protocol) then return top..link.."|h["..displaytext..bot; end
	link=link.."."..protocol; if (uselink) then displaytext=link; end
	if (not toon) then return top..link.."|h["..displaytext..bot; end
	link=link.."@"..toon..location; if (uselink) then displaytext=link; end
	return top..link.."|h["..displaytext..bot;
end

-- An event has been received
--function wn:OnEvent(event,arg1,arg2,arg3,arg4,_,_,_,_,arg9)
function wn:OnEvent(event,...)
	if (wn[event]) then return wn[event](wn,...); end
end

--arg1	chat message
--arg2	author
--arg8	channel number
--arg9	channel name without number (this is _sometimes_ in lowercase)
-- WoWnet public access
function wn:CHAT_MSG_CHANNEL(arg1,arg2,arg3,arg4,_,_,_,_,arg9)
	if (arg9:sub(1,6):lower()~="wownet") then return; end		-- not wownet
	if (pub.Sent[arg1]) then pub.Sent[arg1]=nil; return; end	-- my own output
	-- incoming public data
	local add,prot,ep,data=pub:SplitInput(arg1);
	if (not add or not prot or not ep or not data) then return; end	-- something else
	ep=tonumber(ep);
	-- public broadcast
	table.insert(pub.Incoming,{ ad=add, pr=prot, to=arg2, da=data, ep=ep } );
end

--arg1	"YOU_JOINED", "YOU_LEFT", "THROTTLED"
--arg9	channel name without number (this is _sometimes_ in lowercase)
-- system channel information
function wn:CHAT_MSG_CHANNEL_NOTICE(arg1,_,_,_,_,_,_,_,arg9)
	if (arg9:sub(1,6):lower()~="wownet") then return; end		-- not wownet
	if (arg1=="YOU_LEFT") then
		wn.Connected=nil;
		DM:Chat("Diconnected from WoWnet channel. Reconnecting...");
		wn:Connect(nil,true);
	end
	-- when this happens we must assume the server is ready for our channel
	if (arg1=="YOU_JOINED") then
--print(arg1,arg9);
		if (wn.Connected) then return; end
		wnBoot=time()+4;						-- set connection timer
	end
end

function wn:Connect(cbSR,nocreate)
	DM.MT:Run("wn:Connect",wn.MT.Connect_,wn.MT);
	if (nocreate) then return; end
	return wn:Public(cbSR);
end

function wn.MT:Connect_()
	if (wn.Connected) then return; end					-- done already
	while(time()<wnBoot) do DM.MT:Yield(true); end		-- waiting for boot timer
	local channel="WoWnet"
	JoinTemporaryChannel(channel);
	DM.MT:Yield(true);
	wn.Connected=GetChannelName(channel);
	if (not wn.Connected) then
		DM:Chat("Regular WoWnet channel unavailable on this server.");
		DM:Chat("Connecting to fallback channel...");
		local count=0;
		while(not wn.Connected) do
			count=count+1;
			channel="WoWnet"..count;
			JoinTemporaryChannel(channel);
			DM.MT:Yield(true);
			wn.Connected=GetChannelName(channel);
		end
		DM:Chat("Available WoWnet channel found.");
	else
		DM:Chat("Connected to WoWnet.");
	end
end
