--[[
	MessageBoard - A World of Warcraft in-game forum
	Copyright (C) 2015  Darrell Keys

	This file is part of MessageBoard.

	MessageBoard is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	MessageBoard is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with MessageBoard.  If not, see <http://www.gnu.org/licenses/>.
]]
local addonName, PRIVATE = ... -- get our addon scoped table
local MBFunc = PRIVATE.Functions -- and a reference to our utility functions storage
local MBMACRO_ICON_FILENAMES = {};
local Net=DuckLibMB.Net;
local DB=DuckLibMB.Database;
local READY_CHECK_WAITING_TEXTURE = READY_CHECK_WAITING_TEXTURE or "Interface\\RaidFrame\\ReadyCheck-Waiting";
local READY_CHECK_READY_TEXTURE = READY_CHECK_READY_TEXTURE or "Interface\\RaidFrame\\ReadyCheck-Ready";
local READY_CHECK_NOT_READY_TEXTURE = READY_CHECK_NOT_READY_TEXTURE or "Interface\\RaidFrame\\ReadyCheck-NotReady";
local READY_CHECK_AFK_TEXTURE = READY_CHECK_AFK_TEXTURE or "Interface\\RaidFrame\\ReadyCheck-NotReady";
local L = LibStub("AceLocale-3.0"):GetLocale("MessageBoard")

function MessageBoardXML:ShowMainWindow()
	MessageBoard_OFrame_OpenAvatar:SetText(L["OPT_BUTTON_AVATAR"])
	MessageBoard_OFrame_OpenCategories:SetText(L["OPT_BUTTON_CATEGORY"])
	MessageBoard_OFrame_OpenRanks:SetText(L["OPT_BUTTON_RANK"])
	MessageBoard_OFrame_OpenCheckVersions:SetText(L["OPT_BUTTON_VERSIONS"])
	local _,avatar=MessageBoardXML:ValidateBoardDB(); if (not avatar) then return nil; end
	local class, classFileName = UnitClass("Player")
	local color = RAID_CLASS_COLORS[classFileName]
	local name=UnitName("player");
	name = Ambiguate(name,"none")
	local coloredname="|c"..color.colorStr..name.."|r";
	local pic=nil;
	MessageBoardXML:HideSettings();
	MessageBoard_OFrame_UserSettings:Show();
	MessageBoard_OFrame:Show();
	MessageBoard_OFrameTitle:Show();
	MessageBoard_OFrameTitle:SetText(L["OPT_USER_TITLE"]);
	MessageBoard_OFrame_UserSettings_UserName:SetText(coloredname);
	MessageBoard_OFrame_UserSettings_PostCount:SetText(L["OPT_POST_COUNT"])
	MessageBoard_OFrame_UserSettings_ForumRank:SetText(L["OPT_FORUM_RANK"])
	MessageBoard_OFrame_UserSettings_Save:SetText(L["OPT_BUTTON_SAVE"])
	MessageBoard_OFrame_UserSettings_PostCountText:SetText(MessageBoardXML:UserMessageCount(name))
	MessageBoard_OFrame_UserSettings_ForumRankText:SetText(MessageBoardXML.MyRank);
	MessageBoard_OFrame_UserSettings_Notify:SetChecked(MessageBoard_LS.NewPostNotify)
	MessageBoard_OFrame_UserSettings_NotifyText:SetText(L["OPT_USER_CHATNOTIFY"])
	MessageBoard_OFrame_UserSettings_NotifyPopup:SetChecked(MessageBoard_LS.NewPostNotifyPopup)
	MessageBoard_OFrame_UserSettings_NotifyPopupText:SetText(L["OPT_USER_POPUPNOTIFY"])
	MessageBoard_OFrame_UserSettings_MinimapButton:SetChecked(MessageBoardLDBIconDB.hide);
	MessageBoard_OFrame_UserSettings_MinimapButtonText:SetText(L["OPT_USER_MINIMAP"])
	MessageBoard_OFrame_UserSettings_Quoted:SetChecked(MessageBoard_LS.Quoted);
	MessageBoard_OFrame_UserSettings_QuotedText:SetText(L["OPT_USER_QUOTETEXT"])
	MessageBoard_OFrameBG:SetTexture(MessageBoard_LS.OptBackground); -- set options bg
	if (avatar[name]) then pic=avatar[name]; else pic="Interface\\Icons\\INV_Misc_QuestionMark"; end
	MessageBoard_OFrame_UserSettings_CurrentIcon:SetTexture(pic);
	MessageBoard_OFrame_UserSettings_BGDownButtonBGName:SetText(MessageBoard_LS.BackgroundName)
	MessageBoard_OFrame_UserSettings_BGDownButtonBGTitle:SetText(L["OPT_USER_BGLABEL"])
	if MessageBoardXML.Administrator then
		MessageBoard_OFrame_OpenCategories:Show();
		MessageBoard_OFrame_OpenRanks:Show();
		MessageBoard_OFrame_OpenCheckVersions:Show();
	end
end

function MessageBoardXML:SaveUserSettings()
	if (_G["MessageBoard_OFrame_UserSettings_Notify"]:GetChecked()) then MessageBoard_LS.NewPostNotify=true; else MessageBoard_LS.NewPostNotify=false; end
	if (_G["MessageBoard_OFrame_UserSettings_NotifyPopup"]:GetChecked()) then MessageBoard_LS.NewPostNotifyPopup=true; else MessageBoard_LS.NewPostNotifyPopup=false; end
    if (_G["MessageBoard_OFrame_UserSettings_Quoted"]:GetChecked()) then MessageBoard_LS.Quoted=true; else MessageBoard_LS.Quoted=false; end
	if (_G["MessageBoard_OFrame_UserSettings_MinimapButton"]:GetChecked()) then MessageBoardLDBIconDB.hide=true; else MessageBoardLDBIconDB.hide=false; end
	DuckLibMB:Chat("User Settings Saved.",0,1,0);
	MessageBoardXML:Update();
end

function MessageBoardXML:ShowAvatarWindow()
	MessageBoard_OFrameTitle:SetText(L["OPT_AV_TITLE"]);
	MessageBoard_OFrame_Avatar_Save:SetText(L["OPT_BUTTON_SAVE"])
	MessageBoard_OFrame_Avatar_ClearIcon:SetText(L["OPT_BUTTON_CLEAR"])
	MessageBoard_OFrame_AvatarAvatarListTitle:SetText(L["OPT_AV_AVAILABLE"])
	MessageBoardXML:HideSettings();
	MessageBoard_OFrame_Avatar:Show();
	self:ListAvailableAvatars();
	local _,avatar=self:ValidateBoardDB();
	local name=UnitName("player");
	name = Ambiguate(name,"none")
	if (not avatar or not avatar[name]) then return; end
	MessageBoard_OFrame_AvatarSel_CurrentIcon:SetTexture(avatar[name]);
end
 
function MessageBoardXML:SaveAvatar()
	local _,avatar=self:ValidateBoardDB(); if (not avatar) then return; end
	local name=UnitName("player");
	name = Ambiguate(name,"none")
	avatar[name]=MessageBoard_OFrame_AvatarSel_CurrentIcon:GetTexture();
	--DuckLibMB:Chat("Avatar Saved");
	MessageBoard_OFrame_Avatar:Hide();
	MessageBoardXML:ShowMainWindow();
	MessageBoardXML:Update();
end
 
function MessageBoardXML:ShowAdminCatWindow()
	local f = CreateFrame("Frame",nil,MessageBoard_OFrame)
	f:SetFrameStrata("BACKGROUND")
	f:SetWidth(428) -- Set these to whatever height/width is needed 
	f:SetHeight(464) -- for your Texture

	local t = f:CreateTexture(nil,"ARTWORK")
	t:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Factions.blp")
	t:SetAllPoints(f)
	f.texture = t

	f:SetPoint("CENTER",0,0)
	f:Show()
	MessageBoard_OFrameTitle:SetText(L["OPT_CAT_TITLE"])
	MessageBoard_OFrame_Categories_Save:SetText(L["OPT_BUTTON_SAVE"])
	MessageBoard_OFrame_CategoriesIconTitle:SetText(L["OPT_CAT_ICONLABEL"])
	MessageBoard_OFrame_CategoriesOfficersTitle:SetText(L["OPT_CAT_OFFICERLABEL"])
	MessageBoard_OFrame_CategoriesNormalsTitle:SetText(L["OPT_CAT_NORMALLABEL"])
	MessageBoard_OFrame_CategoriesCategoryOfficerReadTitle:SetText(L["OPT_CAT_READLABEL"])
	MessageBoard_OFrame_CategoriesCategoryOfficerWriteTitle:SetText(L["OPT_CAT_WRITELABEL"])
	MessageBoard_OFrame_CategoriesCategoryUserReadTitle:SetText(L["OPT_CAT_READLABEL"])
	MessageBoard_OFrame_CategoriesCategoryUserWriteTitle:SetText(L["OPT_CAT_WRITELABEL"])
	MessageBoardXML:HideSettings()
	MessageBoard_OFrame_Categories:Show()
	local settings=MessageBoardXML:ValidateBoard().Settings
	local index
	for index=1,9 do
		_G["MessageBoard_OFrame_Categories_Category"..index.."ValueText"]:SetText(L["OPT_CAT_LABEL"]..index)
		if MessageBoardXML:Setting("Category"..index.."Value") then _G["MessageBoard_OFrame_Categories_Category"..index.."Value"]:SetText(MessageBoardXML:Setting("Category"..index.."Value")); end
	end
	for index=1,9 do
		_G["MessageBoard_OFrame_Categories_Category"..index.."OfficerRead"]:SetChecked(MessageBoardXML:Setting("Category"..index.."OfficerRead"));
		_G["MessageBoard_OFrame_Categories_Category"..index.."OfficerWrite"]:SetChecked(MessageBoardXML:Setting("Category"..index.."OfficerWrite"));
		_G["MessageBoard_OFrame_Categories_Category"..index.."UserRead"]:SetChecked(MessageBoardXML:Setting("Category"..index.."UserRead"));
		_G["MessageBoard_OFrame_Categories_Category"..index.."UserWrite"]:SetChecked(MessageBoardXML:Setting("Category"..index.."UserWrite"));
		_G["MessageBoard_OFrame_Categories_Current"..index.."Icon"]:SetTexture(MessageBoardXML:Setting("Category"..index.."Icon"));
	end
end

function MessageBoardXML:SaveAdminCatSettings()
	local settings=MessageBoardXML:ValidateBoard().Settings;
    local buf;
    if (type(settings)=="database") then buf=totable(settings);
    elseif (type(settings)=="string") then buf=unpacktable(settings);
    else buf={}; end
	local index;
	for index=1,9 do
		buf["Category"..index.."Value"]=_G["MessageBoard_OFrame_Categories_Category"..index.."Value"]:GetText();
    end
    for index=1,9 do
		buf["Category"..index.."OfficerRead"]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Categories_Category"..index.."OfficerRead"]:GetChecked());
        buf["Category"..index.."OfficerWrite"]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Categories_Category"..index.."OfficerWrite"]:GetChecked());
        buf["Category"..index.."UserRead"]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Categories_Category"..index.."UserRead"]:GetChecked());
        buf["Category"..index.."UserWrite"]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Categories_Category"..index.."UserWrite"]:GetChecked());
    end
	MessageBoardXML:ValidateBoard().Settings=packtable(buf);
	MessageBoardXML:ShowMainWindow();
	MessageBoardXML:CheckCategories();
	MessageBoardXML:Update();
end

function MessageBoardXML:ShowAdminRanksWindow()
	MessageBoard_OFrameTitle:SetText(L["OPT_RANK_TITLE"]);
	MessageBoard_OFrame_Ranks_Save:SetText(L["OPT_BUTTON_SAVE"])
	MessageBoard_OFrame_RanksDescription:SetText(L["OPT_RANK_HELP1"])
	MessageBoard_OFrame_RanksDescription2:SetText(L["OPT_RANK_HELP2"])
	MessageBoard_OFrame_RanksOfficerPermissionsListTitle:SetText(L["OPT_RANK_OFFICER"])
	MessageBoard_OFrame_RanksNormalPermissionsListTitle:SetText(L["OPT_RANK_NORMAL"])
	MessageBoard_OFrame_RanksAltPermissionsListTitle:SetText(L["OPT_RANK_ALT"])
	MessageBoard_OFrame_RanksNoPostPermissionsListTitle:SetText(L["OPT_RANK_NOACCESS"])
	MessageBoardXML:HideSettings();	MessageBoard_OFrame_Ranks:Show();
	local settings=MessageBoardXML:ValidateBoard().Settings;
	local index;
		for index=2,9 do
		_G["MessageBoard_OFrame_Ranks_OfficerRank"..index.."Text"]:SetText(GuildControlGetRankName(index));
		_G["MessageBoard_OFrame_Ranks_NormalRank"..index.."Text"]:SetText(GuildControlGetRankName(index));
		_G["MessageBoard_OFrame_Ranks_AltRank"..index.."Text"]:SetText(GuildControlGetRankName(index));
		_G["MessageBoard_OFrame_Ranks_NoPostRank"..index.."Text"]:SetText(GuildControlGetRankName(index));
		if (settings) then
			_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]:SetChecked(MessageBoardXML:Setting("OfficerRank"..index));
			_G["MessageBoard_OFrame_Ranks_NormalRank"..index]:SetChecked(MessageBoardXML:Setting("NormalRank"..index));
			_G["MessageBoard_OFrame_Ranks_AltRank"..index]:SetChecked(MessageBoardXML:Setting("AltRank"..index));
			_G["MessageBoard_OFrame_Ranks_NoPostRank"..index]:SetChecked(MessageBoardXML:Setting("NoPostRank"..index));
		end
	end
end

function MessageBoardXML:SaveAdminRankSettings()
	local settings=MessageBoardXML:ValidateBoard().Settings;
    local buf;
    if (type(settings)=="database") then buf=totable(settings);
    elseif (type(settings)=="string") then buf=unpacktable(settings);
    else buf={}; end
	local index;
	for index=2,9 do
		buf["OfficerRank"..index]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]:GetChecked());
		buf["NormalRank"..index]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Ranks_NormalRank"..index]:GetChecked());
        buf["AltRank"..index]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Ranks_AltRank"..index]:GetChecked());
        buf["NoPostRank"..index]=MessageBoardXML:MakeBool(_G["MessageBoard_OFrame_Ranks_NoPostRank"..index]:GetChecked());
    end
	MessageBoardXML:ValidateBoard().Settings=packtable(buf);
	DuckLibMB:Chat("Rank Permissions Saved.",0,1,0);
	MessageBoardXML:ShowMainWindow();
	MessageBoardXML:Update();
end

function MessageBoardXML:SelectIcon(self)
	 MessageBoard_OFrame_AvatarSel_CurrentIcon:SetTexture(_G[self:GetName().."Icon"]:GetTexture());
end

function MessageBoardXML:SelectCatIcon(self,catidx)
	local menu=DMMenuCreate(self)
	menu.Timer=0.2    -- value in seconds
	--menu:Add("Select an Icon", "", DDT_TEXT)
	menu:Header("Select an Icon")
    menu:Separator()
	menu:Button("Anvil",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Anvil") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Anvil")
	menu:Button("Barrel",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\barrel") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\barrel")
	menu:Button("Letter",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Letter") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Letter")
	menu:Button("Mount",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Mount") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Mount")
	menu:Button("Poster",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Poster") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Poster")
	menu:Button("Shield",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Shield") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Shield")
	menu:Button("Chest",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\chest2") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\chest2")
	menu:Button("Open Chest",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Chest") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Chest")
	menu:Button("Staff Head",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Staffhead") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Staffhead")
	menu:Button("Alliance Banner",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\AllianceBanner") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\AllianceBanner")
	menu:Button("Horde Banner",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\HordeBanner") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\HordeBanner")
	menu:Button("Ink Book",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Inkbook") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Inkbook")
	menu:Button("Keg",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\keg") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\keg")
	menu:Button("Cauldron",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\cauldron") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\cauldron")
	menu:Button("Skull",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Skull") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Skull")
	menu:Button("Mail",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\Mail") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\Mail")
	menu:Button("Gold Coin",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\goldicon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\goldicon")
	menu:Button("Wolf",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\wolficon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\wolficon")	
	menu:Button("Fireball",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\fireballicon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\fireballicon")	
	menu:Button("Hearthstone",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\hsicon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\hsicon")
	menu:Button("Diablo III",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\d3icon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\d3icon")
	menu:Button("Starcraft 2",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\sc2icon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\sc2icon")
	menu:Button("Heroes of the Storm",function () MessageBoardXML:SetCatIcon(catidx,"Interface\\AddOns\\MessageBoard\\Images\\hotsicon") end,nil,nil,nil,"Interface\\AddOns\\MessageBoard\\Images\\hotsicon")
	menu:Show()
end

function MessageBoardXML:SelectCatIconOnHide(self)
	MB_DropDown = MB_DropDown and MB_DropDown:Release()
end
function MessageBoardXML:SetCatIcon(self,icon)
	local settings=MessageBoardXML:ValidateBoard().Settings;
    local buf;
    if (type(settings)=="database") then buf=totable(settings);
    elseif (type(settings)=="string") then buf=unpacktable(settings);
    else buf={}; end
	_G["MessageBoard_OFrame_Categories_Current"..self.."Icon"]:SetTexture(icon);
	_G["MessageBoard_MainNavBarButton"..self.."Icon"]:SetTexture(icon);
	buf["Category"..self.."Icon"]=icon;
	MessageBoardXML:ValidateBoard().Settings=packtable(buf);
	DuckLibMB:Chat("Icon Saved.",0,1,0);
end

			
function MBFunc.MBGetSpellorMacroIconInfo(index)
	if ( not index ) then
		return;
	end
	return MBMACRO_ICON_FILENAMES[index];
end
function MBFunc.RefreshPlayerSpellIconInfo()
	MBMACRO_ICON_FILENAMES = {};
	MBMACRO_ICON_FILENAMES[1] = "INV_MISC_QUESTIONMARK";
	local index = 2;
	local numFlyouts = 0;	
	for i = 1, GetNumSpellTabs() do
		local tab, tabTex, offset, numSpells, _ = GetSpellTabInfo(i);
		offset = offset + 1;
		local tabEnd = offset + numSpells;
		for j = offset, tabEnd - 1 do
			--to get spell info by slot, you have to pass in a pet argument
			local spellType, ID = GetSpellBookItemInfo(j, "player"); 
			if (spellType ~= "FUTURESPELL") then
				local spellTexture = strupper(GetSpellBookItemTexture(j, "player"));
				if ( not string.match( spellTexture, "INTERFACE\\BUTTONS\\") ) then
					MBMACRO_ICON_FILENAMES[index] = gsub( spellTexture, "INTERFACE\\ICONS\\", "");
					index = index + 1;
				end
			end
			if (spellType == "FLYOUT") then
				local _, _, numSlots, isKnown = GetFlyoutInfo(ID);
				if (isKnown and numSlots > 0) then
					for k = 1, numSlots do 
						local spellID, overrideSpellID, isKnown = GetFlyoutSlotInfo(ID, k)
						if (isKnown) then
							MBMACRO_ICON_FILENAMES[index] = gsub( strupper(GetSpellTexture(spellID)), "INTERFACE\\ICONS\\", ""); 
							index = index + 1;
						end
					end
				end
			end
		end
	end
	GetMacroIcons( MBMACRO_ICON_FILENAMES );
	GetMacroItemIcons( MBMACRO_ICON_FILENAMES );
end
			
function MessageBoard_OFrame_AvatarSlider_OnValueChanged(self, value)
   -- start fix
   if not self._onsetting then   -- is single threaded 
     self._onsetting = true
     self:SetValue(self:GetValue())
     value = self:GetValue()     -- cant use original 'value' parameter
     self._onsetting = false
   else return end               -- ignore recursion for actual event handler
   -- end fix
   _G[self:GetName().."Text"]:SetText(value) -- handle the event
 end
			
function MessageBoardXML:ListAvailableAvatars()
	local icon,button;
	local index;
	local max=0;
	local texture;
	MBFunc.RefreshPlayerSpellIconInfo();
	local numMacroIcons = #MBMACRO_ICON_FILENAMES;
	if (numMacroIcons<=49) then max=0; else max=math.ceil(numMacroIcons)-14; end
	MessageBoard_OFrame_AvatarSlider:SetMinMaxValues(0,max);
	local offset=MessageBoard_OFrame_AvatarSlider:GetValue();
	-- Icon list
    for i=1,49 do
        icon=_G["MessageBoard_OFrame_AvatarSel_"..i.."Icon"];
        button=_G["MessageBoard_OFrame_AvatarSel_"..i];
        if (not icon or not button) then return; end
        index=i+offset;
		texture = MBFunc.MBGetSpellorMacroIconInfo(index);
		if (index<=numMacroIcons and texture) then
			icon:SetTexture("Interface\\Icons\\"..texture);
            button:Show();
        else
            button:Hide();
        end
    end 
end

function MessageBoardXML:CheckClick(widget)
    local index=2;
    while (_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]) do
        if (widget==_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]) then
			_G["MessageBoard_OFrame_Ranks_NormalRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_AltRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_NoPostRank"..index]:SetChecked(false);
        end
		if (widget==_G["MessageBoard_OFrame_Ranks_NormalRank"..index]) then
			_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_AltRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_NoPostRank"..index]:SetChecked(false);
        end
        if (widget==_G["MessageBoard_OFrame_Ranks_AltRank"..index]) then
			_G["MessageBoard_OFrame_Ranks_NormalRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_OfficerRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_NoPostRank"..index]:SetChecked(false);
        end
        if (widget==_G["MessageBoard_OFrame_Ranks_NoPostRank"..index]) then
			_G["MessageBoard_OFrame_Ranks_OfficerRank"..index]:SetChecked(false);
			_G["MessageBoard_OFrame_Ranks_NormalRank"..index]:SetChecked(false);
            _G["MessageBoard_OFrame_Ranks_AltRank"..index]:SetChecked(false);
        end
		index=index+1;
    end
end

function MessageBoardXML:HideSettings()
	MessageBoard_OFrame_UserSettings:Hide();
	MessageBoard_OFrame_Ranks:Hide();
	MessageBoard_OFrame_Categories:Hide();
	MessageBoard_OFrame_Avatar:Hide();
end

function MessageBoardXML:SelectBG(self)
	local menu=DMMenuCreate(self)
	menu.Timer=0.2;    -- value in seconds
	menu:Add("Classic",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-CLASSIC","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-CLASSIC","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-CLASSIC","Classic") end)
	menu:Add("The Burning Crusade",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-BC","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-BC","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-BC","The Burning Crusade") end)
	menu:Add("Wrath of the Lich King",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-WRATH","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-WRATH","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-WOTLK","Wrath of the Lich King") end)
	menu:Add("Cataclysm",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-CATA","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-CATA","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-CATA","Cataclysm") end)
	menu:Add("Mists of Pandaria",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-MOP","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-MOP","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-MOP","Mists of Pandaria") end)
	menu:Add("Warlords of Draenor",function () MessageBoardXML:SetBG(self,"Interface\\AddOns\\MessageBoard\\Images\\MessageBoardBG-WOD","Interface\\AddOns\\MessageBoard\\Images\\MessageBoard_Options-WOD","Interface\\AddOns\\MessageBoard\\Images\\buttonbg-WOD","Warlords of Draenor") end)
	menu:Show()
end

function MessageBoardXML:SetBG(self,bg,optbg,buttonbg,title)
	MessageBoard_MainLowerFrameBG:SetTexture(bg)
	MessageBoard_OFrameBG:SetTexture(optbg)
	MessageBoard_MainNavBarButton1BG:SetTexture(buttonbg)
	for i=1,9 do
		_G["MessageBoard_MainNavBarButton"..i.."BG"]:SetTexture(buttonbg)
	end
	MessageBoard_OFrame_UserSettings_BGDownButtonBGName:SetText(title)
	MessageBoard_LS.Background=bg
	MessageBoard_LS.BackgroundName=title
	MessageBoard_LS.OptBackground=optbg
	MessageBoard_LS.buttonbg=buttonbg
end

-- To create a menu:
    -- local menu=DMMenuCreate(parent)
-- All your further access to the menu is through this variable.
-- You don't need a parent widget. If you do provide one, the menu will hang down from it.

-- When you're done building your menu, you show it like this:
    -- menu:Show()

-- There are several ways to visually change the menu, but get it working first.

-- To add a header:
    -- menu:Header(text)

-- To add an empty space:
    -- menu:Separator()

-- To add static text:
    -- menu:Text(text)
-- Because I haven't finished this one properly, it will do the same as "Header" for now. Meaning blue text.

-- To add a standard clicker:
    -- menu:Button(text,func,tooltip,inactive,indent,icon,group)
-- Everything except "text" is optional, even though it doesn't make sense to have a clicker without a function for it to call.

-- To add a checkbox:
    -- menu:Checkbox(text,func,checked,tooltip,inactive,indent,icon,group)

-- To add a radiobutton:
    -- menu:Radiobutton(text,func,checked,tooltip,inactive,indent,icon,group)

-- checked - boolean to visually display as checked or unchecked.
-- tooltip - Either a tring used for mouse-over tooltip, or a list of strings where each string is a new line: {"line 1","line 2"}
-- inactive - inhibits user interaction with the entry.
-- indent - a visual indent in steps (not pixels)
-- icon - the rightmost icon. If supplied for a submenu, it will override the default submenu icon.
-- group - group for radiobutton.

-- Everything is optional.

-- To add a submenu:
    -- local submenu=menu:Submenu(text)

-- The return value from this should be used exactly as any other menu, where you can/must add your stuff to it.

-- To add an editbox to another entry, you should for clarity's sake grab the handle first:
    -- local entry=menu:Button("Clickaroo",function(button,checked,edited) end)
    -- entry:AddEdit(text,size,unit,lower,upper)

-- The parameters here are:
    -- text - edit-contents when opening menu
    -- size - width of box in WoW-pixels
    -- unit - "unit" in the DropCount image is "min".
    -- lower/upper - limits for numeric entries